
/*****************************************************************************
 * gcrLib : Graphics Rocchini's Library V0.9 (the nth OpenGL wrapper)        *
 *					                                                         *
 * Copyright(C) 1999 by Claudio Rocchini. All rights reserved.               *
 *					                                                         *
 * Permission  to use, copy, modify, distribute  and sell this  software and *
 * its documentation for any purpose is hereby granted without fee, provided *
 * that  the above copyright notice appear  in all copies and that both that *
 * copyright   notice  and  this  permission  notice  appear  in  supporting *
 * documentation. the author makes  no representations about the suitability *
 * of this software for any purpose. It is provided  "as is" without express *
 * or implied warranty.                                                      *
 *					                                                         *
 *****************************************************************************/

#pragma warning( disable : 4514 4201 )

#include <math.h>
#include <stdio.h>
#include <time.h>

#include "coord3f.h"
#include "viewer.h"


/*
 * This size should really be based on the distance from the center of
 * rotation to the point on the object underneath the mouse.  That
 * point would then track the mouse as closely as possible.  This is a
 * simple example, though, so that is left as an Exercise for the
 * Programmer.
 */

static const float TRACKBALLSIZE = 0.8f;

//#define DELTASPIN		8


	// Project an x,y pair onto a sphere of radius r OR a hyperbolic sheet
	// if we are away from the center of the sphere.
static float TBProjectToSphere(float r, float x, float y)
{
    float d, t, z;

    d = sqrtf(x*x + y*y);
    if (d < r * 0.70710678118654752440f) {    /* Inside sphere */
	z = sqrtf(r*r - d*d);
    } else {           /* On hyperbola */
        t = r / 1.41421356237309504880f;
        z = t*t / d;
    }
    return z;
}

/*
 * Quaternions always obey:  a^2 + b^2 + c^2 + d^2 = 1.0
 * If they don't add up to 1.0, dividing by their magnitued will
 * renormalize them.
 *
 * Note: See the following for more information on quaternions:
 * 
 * - Shoemake, K., Animating rotation with quaternion curves, Computer
 *   Graphics 19, No 3 (Proc. SIGGRAPH'85), 245-254, 1985.
 * - Pletinckx, D., Quaternion calculus as a basic tool in computer
 *   graphics, The Visual Computer 5, 2-13, 1989.
 */
static void NormalizeQuat(float q[4])
{
    int i;
    float mag;

    mag = sqrtf(q[0]*q[0] + q[1]*q[1] + q[2]*q[2] + q[3]*q[3]);
    for (i = 0; i < 4; i++) q[i] /= mag;
}


/*
 *  Given an axis and angle, compute quaternion.
 */
static void AxisToQuat(Coord3f & a, float phi, float q[4])
{
	float s = sinf(phi/2.0f);

    a.Normalize();
	q[0] = a.x*s;
	q[1] = a.y*s;
	q[2] = a.z*s;
    q[3] = cosf(phi/2.0f);
}


/*
 * Ok, simulate a track-ball.  Project the points onto the virtual
 * trackball, then figure out the axis of rotation, which is the cross
 * product of P1 P2 and O P1 (O is the center of the ball, 0,0,0)
 * Note:  This is a deformed trackball-- is a trackball in the center,
 * but is deformed into a hyperbolic sheet of rotation away from the
 * center.  This particular function was chosen after trying out
 * several variations.
 * 
 * It is assumed that the arguments to this routine are in the range
 * (-1.0 ... 1.0)
 */
static void CalcQuat(float q[4], float p1x, float p1y, float p2x, float p2y)
{
    Coord3f a; /* Axis of rotation */
    float phi;  /* how much to rotate about axis */
    Coord3f p1,p2,d;
    float t;

    if (p1x == p2x && p1y == p2y)
	{
	    q[0] = 0.0f;
		q[1] = 0.0f;
		q[2] = 0.0f;
		q[3] = 1.0f; 
        return;
    }

    /*
     * First, figure out z-coordinates for projection of P1 and P2 to
     * deformed sphere
     */
    p1.Set(p1x,p1y,TBProjectToSphere(TRACKBALLSIZE,p1x,p1y));
    p2.Set(p2x,p2y,TBProjectToSphere(TRACKBALLSIZE,p2x,p2y));

    /*
     *  Now, we want the cross product of P1 and P2
     */
	a = p2 ^ p1;

    /*
     *  Figure out how much to rotate around that axis.
     */
	d = p1 - p2;
    t = d.Norm() / (2.0f*TRACKBALLSIZE);

    /*
     * Avoid problems with out-of-control values...
     */
    if (t > 1.0f) t = 1.0f;
    if (t < -1.0f) t = -1.0f;
    phi = 2.0f * asinf(t);

    AxisToQuat(a,phi,q);
}


/*
 * Given two rotations, e1 and e2, expressed as quaternion rotations,
 * figure out the equivalent single rotation and stuff it into dest.
 * 
 * This routine also normalizes the result every RENORMCOUNT times it is
 * called, to keep error from creeping in.
 *
 * NOTE: This routine is written so that q1 or q2 may be the same
 * as dest (or each other).
 */

static void AddQuats(float q1[4], float q2[4], float dest[4])
{
	const int RENORMCOUNT = 97;
    static int count=0;
	Coord3f t1,t2,t3,tf;
    float d;

	t1.x = q1[0];
	t1.y = q1[1];
	t1.z = q1[2];

	t2.x = q2[0];
	t2.y = q2[1];
	t2.z = q2[2];

	d  = t2 * t1;
	t3 = t2 ^ t1;
	t1 *= q2[3];
	t2 *= q1[3];
	tf = t1 + t2 +t3;

    dest[0] = tf.x;
    dest[1] = tf.y;
    dest[2] = tf.z;
    dest[3] = q1[3] * q2[3] - d;

    if (++count > RENORMCOUNT) {
        count = 0;
        NormalizeQuat(dest);
    }
}

/*
 * Build a rotation matrix, given a quaternion rotation.
 *
 */
static void BuildRotmatrix(float m[4][4], float q[4])
{
    m[0][0] = 1.0f - 2.0f * (q[1] * q[1] + q[2] * q[2]);
    m[0][1] = 2.0f * (q[0] * q[1] - q[2] * q[3]);
    m[0][2] = 2.0f * (q[2] * q[0] + q[1] * q[3]);
    m[0][3] = 0.0f;

    m[1][0] = 2.0f * (q[0] * q[1] + q[2] * q[3]);
    m[1][1] = 1.0f - 2.0f * (q[2] * q[2] + q[0] * q[0]);
    m[1][2] = 2.0f * (q[1] * q[2] - q[0] * q[3]);
    m[1][3] = 0.0f;

    m[2][0] = 2.0f * (q[2] * q[0] - q[1] * q[3]);
    m[2][1] = 2.0f * (q[1] * q[2] + q[0] * q[3]);
    m[2][2] = 1.0f - 2.0f * (q[1] * q[1] + q[0] * q[0]);
    m[2][3] = 0.0f;

    m[3][0] = 0.0f;
    m[3][1] = 0.0f;
    m[3][2] = 0.0f;
    m[3][3] = 1.0f;
}


grViewer::grViewer( void )
{
	Width		= 0;
    Height		= 0;
	MouseDownX	= 0;
	MouseDownY	= 0;
	Spinning	= 0;
	Mode		= NONE;
}



void grViewer::Init( int w, int h )
{
	Spinning	= 0;
    Width  = w;
    Height = h;
    CalcQuat( curquat, 0.0f, 0.0f, 0.0f, 0.0f );
	BuildRotmatrix(mat,curquat);
	scale[0] = 1.0f;
	scale[1] = 1.0f;
	scale[2] = 1.0f;
	trans[0] = 0.0f;
	trans[1] = 0.0f;
	trans[2] = -10.0f;
	rtran[0] = 0.0f;
	rtran[1] = 0.0f;
	rtran[2] = 0.0f;
}


void grViewer::Reset()
{
	Init(Width,Height);
}

void grViewer::Resize( int w, int h )
{
    Width  = w;
    Height = h;
}


void grViewer::StartRotation( int x, int y )
{
	Lastmove    = 0;
	Spinning    = 0;
	MouseDownX	= x;
    MouseDownY	= y;
	Mode		= ROTATION;
}


void grViewer::EndRotation( int /*x*/, int /*y*/ )
{
	if( Spinning && (clock()-Lastmove > CLOCKS_PER_SEC/4) )
		Spinning = 0;
	Mode		= NONE;
}


void grViewer::StartTranslation( int x, int y )
{
	MouseDownX	= x;
    MouseDownY	= y;
	Mode		= TRANSLATION;
}


void grViewer::EndTranslation( int /*x*/, int /*y*/ )
{
	Spinning    = 0;
	Mode		= NONE;
}


void grViewer::StartScale( int x, int y )
{
	MouseDownX	= x;
    MouseDownY	= y;
	Mode		= SCALE;
}


void grViewer::EndScale( int /*x*/, int /*y*/ )
{
	Spinning    = 0;
	Mode		= NONE;
}

void grViewer::End( int x, int y )
{
	switch(Mode)
	{
	case NONE:
		break;
	case ROTATION:
		EndRotation(x,y);
		break;
	case TRANSLATION:
		EndTranslation(x,y);
		break;
	case SCALE:
		EndScale(x,y);
		break;
	}
}

void grViewer::MouseMove( int x, int y )
{
	//const int SPINTH = 3;

	switch(Mode)
	{
	case NONE:
		break;
	case ROTATION:
		Spinning = 0;
			
		if(x!=MouseDownX || y!=MouseDownY)
		{
			CalcQuat(lasquat,
			-(2.0f * (Width-MouseDownX)/Width  - 1.0f),
            -(2.0f *        MouseDownY /Height - 1.0f),
            -(2.0f * (Width-x         )/Width  - 1.0f),
            -(2.0f *        y          /Height - 1.0f)
			);
			AddQuats(lasquat,curquat,curquat);
			BuildRotmatrix(mat,curquat);
			//if( abs(MouseDownX-x)>SPINTH ||
			//	abs(MouseDownY-y)>SPINTH )
			Spinning = 1;
			MouseDownX = x;
			MouseDownY = y;
			Lastmove   = clock();
		}
		break;
	case TRANSLATION:
		if(x != MouseDownX || y != MouseDownY)
		{
			trans[0] -= 4.0f * float(MouseDownX-x)/Width;
			trans[1] += 4.0f * float(MouseDownY-y)/Height;
			MouseDownX = x;
			MouseDownY = y;
		}
		break;
	case SCALE:
		if(y != MouseDownY)
		{
			float d = 2.0f * float(y-MouseDownY)/Height;
			trans[2] += d;
			MouseDownX = x;
			MouseDownY = y;
		}
		break;
	}
}

void grViewer::DoSpinning()
{
	AddQuats(lasquat,curquat,curquat);
	BuildRotmatrix(mat,curquat);
}


